from keras.layers import Input, Dense, Convolution1D, MaxPooling1D, UpSampling1D, Flatten, Lambda, Reshape
from keras.models import Model
from keras import regularizers
import os
from keras import backend as K
import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
from sklearn import cluster, metrics
import time

## using gpu or not
os.environ["CUDA_VISIBLE_DEVICES"]="0" 
config = tf.ConfigProto()
config.gpu_options.allow_growth = True
session = tf.Session(config=config)

# dimensions of our images.
train_data_dir = r'F:\Programming\Python\AE\su_jz for cluster2.csv'
validation_data_dir = r'F:\Programming\Python\AE\su_gsm2016_avg_forCls(test).csv'
output_dir = "F:\Programming\Python\AE\jz_ac\\"

nb_epoch = 300
latent_dim = 4
 
intermediate_dim = 96
original_dim = 168*2
epsilon_std = 1.0
#batch_size = 5 
#batch_size = int(batch_size)

nclusters = 6

# train_data for autoencoder01, autoencoder02
def load_data(filePath):
    train_data01 = []
    train_data02 = []
    with open(filePath,'r+') as f:
        for line in f:
            pData = line.strip().split(",")
            pData = np.reshape(pData, (168, 1)).astype(float)
            p_max = np.max(pData)
            train_data01.append(pData/(p_max+10e-9))
            train_data02.append(pData)
            
    train_data01 = np.array(train_data01) 
    train_data02 = np.array(train_data02)
    
    zmax_train02 = train_data02.max(axis = 0)
    train_data02 = train_data02/(zmax_train02+10e-9)
    
    a = train_data02.max(axis = 1)
    train = [train_data02[i]/(a[i]+10e-9) for i in range(train_data02.shape[0])]
    train = np.reshape(train,(train_data02.shape[0],-1,1))
    
    data = np.hstack((train_data01, train))
    #train_data = train
    return data
      
train_data = load_data(train_data_dir)
validation_data = load_data(validation_data_dir)

kernel_size = 13
original_img_size = (original_dim, 1)
input_shape = Input(original_img_size, name='encoder_input')
x = Convolution1D(24, kernel_size, activation='relu', padding='same', name='conv_1')(input_shape)
x = MaxPooling1D(2, padding='same', name='maxpooling_1')(x)
x = Convolution1D(24, kernel_size, activation='relu', padding='same', name='conv_2')(x)
x = MaxPooling1D(2, padding='same', name='maxpooling_2')(x)
x = Convolution1D(48, 7, activation='relu', padding='same', name='conv_3')(x)
encoded = MaxPooling1D(2, padding='same', name='maxpooling_3')(x)
flat = Flatten( name='flatten_1')(encoded)
hidden = Dense(intermediate_dim, activation='relu', name='dense_1')(flat)

z_mean = Dense(latent_dim, name='z_mean')(hidden)
z_log_var = Dense(latent_dim, name='z_log_var')(hidden)


def sampling(args):
    z_mean, z_log_var = args
    
    epsilon = K.random_normal(shape=(K.shape(z_mean)[0], 4),
                              mean=0., stddev=epsilon_std)
    return z_mean + K.exp(z_log_var/2) * epsilon


z = Lambda(sampling, output_shape=(latent_dim,), name='z')([z_mean, z_log_var])

#decoder_hid = Dense(intermediate_dim)
decoder_hid = Dense(intermediate_dim, activation='relu', name='dense_2')
decoder_upsample = Dense(48 * 42, activation='relu', name='dense_3')
output_shape = (1, 42, 48)
decoder_reshape = Reshape(output_shape[1:], name='reshape_1')

latent_inputs = Input(shape=(latent_dim,))

hid_decoded = decoder_hid(latent_inputs)
up_decoded = decoder_upsample(hid_decoded)
reshape_decoded = decoder_reshape(up_decoded)

x = Convolution1D(48, 7, activation='relu', padding='same', name='conv_4')(reshape_decoded)
x = UpSampling1D(2, name='unsampling_1')(x)
x = Convolution1D(24, kernel_size, activation='relu', padding='same', name='conv_5')(x)
x = UpSampling1D(2, name='unsampling_2')(x)
x = Convolution1D(24, kernel_size, activation='relu', padding='same', name='conv_6')(x)
x = UpSampling1D(2, name='unsampling_3')(x)
decoded = Convolution1D(1, kernel_size, activation='sigmoid', padding='same', name='conv_7')(x)

from keras import objectives
def vae_loss(input_shape, decoded):
    input_shape = K.flatten(input_shape)
    decoded = K.flatten(decoded)
    xent_loss = original_dim * objectives.binary_crossentropy(input_shape, decoded)
    kl_loss = - 0.5 * K.sum(1 + z_log_var - K.square(z_mean) - K.exp(z_log_var), axis=-1)
    
    return xent_loss + kl_loss

encoder = Model(input_shape, [z_mean, z_log_var, z], name='encoder')
decoder = Model(latent_inputs, decoded, name='decoder')
outputs = decoder(encoder(input_shape)[2])
autoencoder = Model(input_shape, outputs, name='vae')

autoencoder.compile(optimizer='adam', loss = vae_loss)
encoder.summary()
decoder.summary()
autoencoder.summary()

time_start=time.time()
hist = autoencoder.fit(
        train_data,train_data,
        shuffle = 'True',
     #   batch_size = batch_size,
        epochs=nb_epoch,
        validation_data=(validation_data, validation_data)
        )
time_end=time.time()
print('totally cost:',time_end-time_start)
autoencoder.save(r'F:\Programming\Python\AE\cov_autoencoder_model_f1f2_h7_3-19.h5')
print('file save success')
print('latent_dim = %d' % latent_dim)
########################################################

## cluster hidden layer
#encoder_1 = Model(input_shape, z_mean)
#decoder_1 = Model(autoencoder.get_layer('lambda_1'), autoencoder.output)


encoder_1_out = encoder.predict(train_data)[0]

# kmeans
kmeans = cluster.KMeans(n_clusters = nclusters)
kmeans.fit(encoder_1_out)
y_cls = kmeans.labels_
y_centers = kmeans.cluster_centers_
np.savetxt('%s%s'%(output_dir, '1_cls.csv'), y_cls,  fmt="%s", delimiter = ',')
np.savetxt('%s%s'%(output_dir,'1_center.csv'), y_centers,  fmt="%s", delimiter = ',')

score = metrics.calinski_harabaz_score(encoder_1_out, y_cls)
score1 = metrics.silhouette_score(encoder_1_out, y_cls)
print("method1 clusters value is %f\t%f\n" % (score, score1))

# spectral
#spectral = cluster.SpectralClustering(n_clusters = nclusters)
#spectral.fit(encoder_1_out)
#y_cls = spectral.labels_.astype(np.int)
#np.savetxt('%s%s'%(output_dir, '1_clsSpectral.csv'), y_cls,  fmt="%s", delimiter = ',')

# circle for kmeans
for n in range(2,20):
    kmeans = cluster.KMeans(n_clusters = n)
    kmeans.fit(encoder_1_out)
    score = metrics.calinski_harabaz_score(encoder_1_out,kmeans.labels_)
    score1 = metrics.silhouette_score(encoder_1_out,kmeans.labels_)
    print("%d clusters kmeans value is %f\t%f" % (n, score, score1))
    
print("-----------------------------")
    
# decode
y_centers = np.array(y_centers)
y_centers_1 = y_centers[:, :]
#decoder_out1 = decoder_1.predict(y_centers_1)
#decoder_out1 = np.reshape(decoder_out1,(-1,168))
#np.savetxt('%s%s'%(output_dir, '1_decoder.csv'), decoder_out1,  fmt="%s", delimiter = ',')
print('file save success\n') 
############################

